/* ownCloud Android Library is available under MIT license
 *   Copyright (C) 2018 ownCloud GmbH.
 *
 *   Permission is hereby granted, free of charge, to any person obtaining a copy
 *   of this software and associated documentation files (the "Software"), to deal
 *   in the Software without restriction, including without limitation the rights
 *   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *   copies of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *
 *   The above copyright notice and this permission notice shall be included in
 *   all copies or substantial portions of the Software.
 *
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *   NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 *   BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 *   ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 *   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *   THE SOFTWARE.
 *
 */

package com.owncloud.android.lib.resources.files;

import com.owncloud.android.lib.common.OwnCloudClient;
import com.owncloud.android.lib.common.http.HttpConstants;
import com.owncloud.android.lib.common.http.webdav.PropfindMethod;
import com.owncloud.android.lib.common.network.WebdavUtils;
import com.owncloud.android.lib.common.operations.RemoteOperation;
import com.owncloud.android.lib.common.operations.RemoteOperationResult;
import com.owncloud.android.lib.common.utils.Log_OC;

import java.util.ArrayList;

import at.bitfire.dav4android.DavResource;
import okhttp3.HttpUrl;
import okhttp3.Response;

import static com.owncloud.android.lib.common.operations.RemoteOperationResult.ResultCode.OK;

/**
 * Remote operation performing the read of remote file or folder in the ownCloud server.
 *
 * @author David A. Velasco
 * @author masensio
 * @author David González Verdugo
 */

public class ReadRemoteFolderOperation extends RemoteOperation {

    private static final String TAG = ReadRemoteFolderOperation.class.getSimpleName();

    private String mRemotePath;

    /**
     * Constructor
     *
     * @param remotePath Remote path of the file.
     */
    public ReadRemoteFolderOperation(String remotePath) {
        mRemotePath = remotePath;
    }

    /**
     * Performs the read operation.
     *
     * @param client Client object to communicate with the remote ownCloud server.
     */
    @Override
    protected RemoteOperationResult run(OwnCloudClient client) {
        RemoteOperationResult result = null;

        try {

            PropfindMethod propfindMethod = new PropfindMethod(
                    client.getOkHttpClient(),
                    HttpUrl.parse(client.getNewWebDavUri() + WebdavUtils.encodePath(mRemotePath)),
                    1);

            Response response = client.executeHttpMethod(propfindMethod);

            boolean isSuccess = (response.code() == HttpConstants.HTTP_MULTI_STATUS
                    || response.code() == HttpConstants.HTTP_OK);

            if (isSuccess) {

                ArrayList<Object> mFolderAndFiles = new ArrayList<>();

                // parse data from remote folder
                mFolderAndFiles.add(
                        new RemoteFile(propfindMethod.getDavResource(), client.getAccount().getDisplayName())
                );

                // loop to update every child
                for (DavResource resource : propfindMethod.getMembers()) {
                    RemoteFile file = new RemoteFile(resource, client.getAccount().getDisplayName());
                    mFolderAndFiles.add(file);
                }

                // Result of the operation
                result = new RemoteOperationResult(OK);
                result.setData(mFolderAndFiles);

            } else {

                // synchronization failed
                result = new RemoteOperationResult(
                        false,
                        propfindMethod.getRequest(),
                        response
                );
            }

        } catch (Exception e) {
            result = new RemoteOperationResult(e);
        } finally {
            if (result.isSuccess()) {
                Log_OC.i(TAG, "Synchronized " + mRemotePath + ": " + result.getLogMessage());
            } else {
                if (result.isException()) {
                    Log_OC.e(TAG, "Synchronized " + mRemotePath + ": " + result.getLogMessage(),
                            result.getException());
                } else {
                    Log_OC.e(TAG, "Synchronized " + mRemotePath + ": " + result.getLogMessage());
                }
            }
        }
        return result;
    }
}